/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.common.callcontext;

import java.util.Optional;
import java.util.UUID;

import jakarta.enterprise.context.ApplicationScoped;
import jakarta.inject.Inject;

import org.apache.logging.log4j.CloseableThreadContext;

import de.ozgcloud.user.common.errorhandling.ResourceNotFoundException;
import io.grpc.ForwardingServerCallListener;
import io.grpc.Metadata;
import io.grpc.ServerCall;
import io.grpc.ServerCall.Listener;
import io.grpc.ServerCallHandler;
import io.grpc.ServerInterceptor;
import io.grpc.Status;
import io.quarkus.grpc.GlobalInterceptor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@ApplicationScoped
@GlobalInterceptor
public class GrpcCallContextInterceptor implements ServerInterceptor {

	private static final String REQUEST_ID_KEY = "requestId";

	static final String KEY_USER_ID = "USER_ID-bin";
	static final String KEY_USER_NAME = "USER_NAME-bin";
	static final String KEY_CLIENT_NAME = "CLIENT_NAME-bin";
	static final String KEY_REQUEST_ID = "REQUEST_ID-bin";
	static final String KEY_ACCESS_LIMITED_ORGAID = "ACCESS_LIMITED_TO_ORGANISATORISCHEEINHEITENID-bin";
	static final String KEY_ACCESS_LIMITED = "ACCESS_LIMITED-bin";

	static final String REQUEST_ID_PLACEHOLDER = "no-requestId-given";

	static final String MISSING_CLIENT_NAME_MESSAGE_TEMPLATE = "Missing client name in grpc header for request %s.";

	@Inject
	CurrentCallContextUserService userService;

	@Override
	public <Q, S> Listener<Q> interceptCall(ServerCall<Q, S> call, Metadata headers, ServerCallHandler<Q, S> next) {
		return new LogContextSettingListener<>(next.startCall(call, headers), call, headers);
	}

	class LogContextSettingListener<A, Q, S> extends ForwardingServerCallListener.SimpleForwardingServerCallListener<A> {
		private final String requestId;
		private final Metadata headers;
		private final ServerCall<Q, S> originCall;
		private final Optional<String> clientName;

		public LogContextSettingListener(ServerCall.Listener<A> delegate, ServerCall<Q, S> originCall, Metadata headers) {
			super(delegate);
			this.headers = headers;
			this.requestId = getRequestId();
			this.originCall = originCall;
			this.clientName = getClientName();
		}

		String getRequestId() {
			return GrpcUtil.getFromHeaders(KEY_REQUEST_ID, headers).orElseGet(() -> UUID.randomUUID().toString());
		}

		private Optional<String> getClientName() {
			return GrpcUtil.getFromHeaders(KEY_CLIENT_NAME, headers);
		}

		@Override
		public void onMessage(A message) {
			doSurroundOn(() -> super.onMessage(message));
		}

		@Override
		public void onHalfClose() {
			doSurroundOn(super::onHalfClose);
		}

		@Override
		public void onCancel() {
			doSurroundOn(super::onCancel);
		}

		@Override
		public void onComplete() {
			doSurroundOn(super::onComplete);
		}

		@Override
		public void onReady() {
			if (clientName.isEmpty()) {
				handleMissingClientName();
				return;
			}
			doSurroundOn(super::onReady);
		}

		void handleMissingClientName() {
			var message = String.format(MISSING_CLIENT_NAME_MESSAGE_TEMPLATE, requestId);
			LOG.error(message);
			originCall.close(Status.PERMISSION_DENIED.withDescription(message), headers);
		}

		void doSurroundOn(Runnable runnable) {
			try (var ctc = CloseableThreadContext.put(REQUEST_ID_KEY, requestId)) {
				startSecurityContext();
				runnable.run();
			} catch (Exception e) {
				handleException(e);
			} finally {
				clearSecurityContext();
			}
		}

		void handleException(Exception e) {
			if (e instanceof ResourceNotFoundException) {
				LOG.warn("Grpc Interceptor: " + e.getMessage(), e);
				originCall.close(Status.NOT_FOUND, headers);
			} else {
				LOG.error("Grpc Interceptor: Unknown exception", e);
				originCall.close(Status.UNKNOWN, headers);
			}
		}

		void startSecurityContext() {
			userService.setCallContextUser(createCallContextUser());
		}

		void clearSecurityContext() {
			userService.clearCallContextUser();
		}

		CallContextUser createCallContextUser() {
			var builder = CallContextUser.builder()
					.clientName(clientName.orElseGet(() -> null))
					.userId(GrpcUtil.getFromHeaders(GrpcCallContextInterceptor.KEY_USER_ID, headers))
					.userName(GrpcUtil.getFromHeaders(KEY_USER_NAME, headers))
					.authenticated(true)
					.organisatorischeEinheitenIds(GrpcUtil.getCollection(KEY_ACCESS_LIMITED_ORGAID, headers));

			GrpcUtil.getFromHeaders(KEY_ACCESS_LIMITED, headers).map(Boolean::parseBoolean).ifPresentOrElse(
					builder::organisationEinheitenIdCheckNecessary,
					() -> builder.organisationEinheitenIdCheckNecessary(false));

			return builder.build();
		}
	}
}
