/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.common.errorhandling;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.UUID;

import jakarta.ws.rs.NotFoundException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;

import com.cronutils.utils.StringUtils;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.user.User;

class ApiErrorUtilTest {

	@Spy
	private final ApiErrorUtil util = new ApiErrorUtil();

	private final String message = LoremIpsum.getInstance().getWords(5);
	private final String exceptionId = UUID.randomUUID().toString();

	@DisplayName("Build from runtime exception")
	@Nested
	class TestBuildFromRuntimeException {

		private final RuntimeException exception = new RuntimeException(message);

		@BeforeEach
		void mockExceptionId() {
			doReturn(exceptionId).when(util).createExceptionId();
		}

		@Test
		void shouldHaveMessageCode() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessageCode()).isEqualTo(ApiErrorUtil.RUNTIME_MESSAGE_CODE);
		}

		@Test
		void shouldHaveMessage() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getMessage());
			assertThat(error.getIssues().get(0).getMessage()).contains(exceptionId);
		}

		@Test
		void shouldHaveExceptionId() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getExceptionId()).isEqualTo(exceptionId);
		}

		private ApiError handleException() {
			return util.buildFromRuntimeException(exception);
		}
	}

	@DisplayName("Build from NotFound exception")
	@Nested
	class TestBuildFromNotFoundException {

		private final NotFoundException exception = new NotFoundException(message);

		@BeforeEach
		void mockExceptionId() {
			doReturn(exceptionId).when(util).createExceptionId();
		}

		@Test
		void shouldHaveMessageCode() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessageCode()).isEqualTo(ApiErrorUtil.RESOURCE_NOT_FOUND_MESSAGE_CODE);
		}

		@Test
		void shouldHaveMessage() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getMessage());
		}

		@Test
		void shouldHaveExceptionId() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getExceptionId()).isEqualTo(exceptionId);
		}

		private ApiError handleException() {
			return util.buildFromNotFoundException(exception);
		}
	}

	@DisplayName("Build from resource not found exception")
	@Nested
	class TestBuildFromResourceNotFoundException {

		private final ResourceNotFoundException exception = new ResourceNotFoundException(User.class, 42L);

		@Test
		void shouldHaveMessageCode() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessageCode()).isEqualTo(ApiErrorUtil.RESOURCE_NOT_FOUND_MESSAGE_CODE);
		}

		@Test
		void shouldHaveMessage() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessage()).isEqualTo(exception.getMessage());
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getExceptionId());
		}

		@Test
		void shouldHaveExceptionId() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getExceptionId()).isEqualTo(exception.getExceptionId());
		}

		private ApiError handleException() {
			return util.buildFromResourceNotFoundException(exception);
		}
	}

	@DisplayName("Build from access forbidden exception")
	@Nested
	class TestHandleAccessDeniedException {

		private final AccessForbiddenException exception = new AccessForbiddenException(User.class, message);

		@BeforeEach
		void mockExceptionId() {
			doReturn(exceptionId).when(util).createExceptionId();
		}

		@Test
		void shouldHaveMessageCode() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessageCode()).isEqualTo(ApiErrorUtil.ACCESS_DENIED_MESSAGE_CODE);
		}

		@Test
		void shouldHaveMessage() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getMessage());
			assertThat(error.getIssues().get(0).getMessage()).contains(exceptionId);
		}

		@Test
		void shouldHaveExceptionId() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getExceptionId()).isEqualTo(exceptionId);
		}

		private ApiError handleException() {
			return util.buildFromAccessForbiddenException(exception);
		}
	}

	@DisplayName("Build from keycloak unavailable exception")
	@Nested
	class TestBuildFromKeycloakUnavailableException {

		private final KeycloakUnavailableException exception = new KeycloakUnavailableException(StringUtils.EMPTY, StringUtils.EMPTY,
				StringUtils.EMPTY, new Throwable());

		@Test
		void shouldHaveMessageCode() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessageCode()).isEqualTo(ApiErrorUtil.SERVICE_UNAVAILABLE_MESSAGE_CODE);
		}

		@Test
		void shouldHaveMessage() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getMessage());
			assertThat(error.getIssues().get(0).getMessage()).contains(exception.getExceptionId());
		}

		@Test
		void shouldHaveExceptionId() {
			var error = handleException();

			assertThat(error.getIssues()).hasSize(1);
			assertThat(error.getIssues().get(0).getExceptionId()).isEqualTo(exception.getExceptionId());
		}

		private ApiError handleException() {
			return util.buildFromKeycloakUnavailableException(exception);
		}
	}

	@Test
	void shouldCreateExceptionId() {
		var exceptionId = util.createExceptionId();

		assertThat(exceptionId).isNotNull();
	}
}